import logging
from smtplib import SMTPException
from typing import Any

from django.conf import settings
from django.contrib.auth import forms as auth_forms
from django.contrib.auth.base_user import BaseUserManager
from django.contrib.auth.forms import AuthenticationForm, UserCreationForm
from django.contrib.auth.forms import PasswordResetForm as ContribPasswordResetForm
from django.contrib.auth.tokens import default_token_generator
from django.contrib.sites.shortcuts import get_current_site
from django.core.exceptions import ValidationError
from django.core.mail import EmailMultiAlternatives
from django.db.models import Q
from django.forms import ChoiceField, EmailField, EmailInput
from django.template import loader
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django.utils.translation import gettext_lazy as _

from core.models import ConferenceMember, PlatformUser, UserCommunicationChannel
from core.tokens import channel_activation_token

logger = logging.getLogger(__name__)


class LoginForm(AuthenticationForm):
    plainui_theme = ChoiceField(required=False, choices=PlatformUser.Theme.choices, label=_('PlainUI Theme'))

    def __init__(self, request, *args, with_theme_switcher=False, **kwargs):
        if not with_theme_switcher and 'plainui_theme' in self.base_fields:
            del self.base_fields['plainui_theme']
        super().__init__(request, *args, **kwargs)

    def clean(self):
        if self.request.limited:
            raise ValidationError(_('login--rate-limited'))
        return super().clean()


class PasswordResetForm(ContribPasswordResetForm):
    def __init__(self, *args: Any, conference=None, **kwargs: Any) -> None:
        self.conference = conference
        super().__init__(*args, **kwargs)

    def get_users(self, email):
        normalized_mail = BaseUserManager.normalize_email(email)
        users = (
            PlatformUser.objects.prefetch_related('communication_channels')
            .filter(communication_channels__address=normalized_mail, communication_channels__channel=UserCommunicationChannel.Channel.MAIL)
            .filter(Q(allow_reset_non_primary=True) | Q(allow_reset_non_primary=False, communication_channels__is_primary=True))
            .distinct()
        )
        for user in users:
            channel_qs = user.communication_channels.filter(address=email)
            channel = channel_qs.first()
            if not auth_forms._unicode_ci_compare(email, channel.address):
                continue
            if not user.has_usable_password():
                continue
            channels = user.communication_channels.filter(channel=UserCommunicationChannel.Channel.MAIL, is_primary=True).union(channel_qs)
            yield (user, channels)

    def save(
        self,
        subject_template_name='registration/password_reset_subject.txt',
        email_template_name='registration/password_reset_email.html',
        use_https=False,
        token_generator=default_token_generator,
        from_email=None,
        request=None,
        html_email_template_name=None,
        extra_email_context=None,
    ):
        """
        Generate a one-use only link for resetting password and send it to the
        user.
        """
        email = self.cleaned_data['email']
        current_site = get_current_site(request)

        for user, channels in self.get_users(email):
            for channel in channels:
                context = {
                    'email': channel.address,
                    'domain': current_site,
                    'site_name': self.conference.name if self.conference else current_site.name,
                    'uid': urlsafe_base64_encode(force_bytes(user.pk)),
                    'user': user,
                    'token': token_generator.make_token(user),
                    'protocol': 'https' if use_https else 'http',
                    **(extra_email_context or {}),
                }
                self.send_mail(
                    subject_template_name,
                    email_template_name,
                    context,
                    from_email,
                    channel.address,
                    html_email_template_name=html_email_template_name,
                )


class RegistrationForm(UserCreationForm):
    class Meta:
        model = PlatformUser
        fields = (
            'username',
            'email',
            'password1',
            'password2',
        )

    email = EmailField(
        label=_('Email'),
        max_length=254,
        required=True,
        help_text=_('Registration__email__help'),
        widget=EmailInput(attrs={'autocomplete': 'email'}),
    )

    def __init__(self, *args: Any, request=None, conference=None, **kwargs: Any) -> None:
        self.request = request
        self.conference = conference
        super().__init__(*args, **kwargs)

    def clean(self):
        if self.request and self.request.limited:
            raise ValidationError(_('Registration__rate-limited'))
        if (username := self.cleaned_data.get('username')) and username.startswith('_'):
            raise ValidationError({'username': _('Registration__username__no_underscore')})
        return super().clean()

    def send_mail(
        self,
        subject_template_name,
        email_template_name,
        context,
        to_email,
        html_email_template_name=None,
    ):
        """
        Send a django.core.mail.EmailMultiAlternatives to `to_email`.
        """
        subject = loader.render_to_string(subject_template_name, context)
        # Email subject *must not* contain newlines
        subject = ''.join(subject.splitlines())
        body = loader.render_to_string(email_template_name, context)

        email_message = EmailMultiAlternatives(
            to=[to_email],
            subject=subject,
            body=body,
            reply_to=settings.MAIL_REPLY_TO,
        )
        if html_email_template_name is not None:
            html_email = loader.render_to_string(html_email_template_name, context)
            email_message.attach_alternative(html_email, 'text/html')

        email_message.send(fail_silently=False)

    def save(
        self,
        activation_url,
        subject_template_name='registration/signup_subject.txt',
        email_template_name='registration/signup_email.html',
        use_https=False,
        token_generator=channel_activation_token,
        request=None,
        html_email_template_name=None,
        extra_email_context=None,
    ):
        """
        Generate a one-use only link for resetting password and send it to the
        user.
        """
        # create user as in_active
        user: PlatformUser = super().save(False)
        user.is_active = False
        user.save()

        # configure communication channel based on given mail
        channel = user.communication_channels.create(
            channel=UserCommunicationChannel.Channel.MAIL,
            address=BaseUserManager.normalize_email(user.email),
        )
        channel.save()

        current_site = get_current_site(request)

        if self.conference and not self.conference.require_ticket:
            cm, _ = ConferenceMember.objects.get_or_create(conference=self.conference, user=user)
            cm.save()

        context = {
            'user': user,
            'email': user.email,
            'domain': current_site,
            'site_name': self.conference.name if self.conference else current_site.name,
            'uid': urlsafe_base64_encode(force_bytes(user.pk)),
            'cid': channel.id,
            'channel': channel,
            'token': token_generator.make_token(channel),
            'protocol': 'https' if use_https else 'http',
            'activation_url': activation_url,
            **(extra_email_context or {}),
        }
        try:
            self.send_mail(
                subject_template_name,
                email_template_name,
                context,
                channel.address,
                html_email_template_name=html_email_template_name,
            )
        except SMTPException as exc:
            user.delete()
            raise exc from None
