import subprocess

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from django.urls import reverse
from django.utils import timezone

from core.models.rooms import Room


def create_hangar(room: Room):
    username = room.assembly.slug
    cmd = ['ssh', '-i', settings.HANGAR_KEY, settings.HANGAR_HOST, settings.HANGAR_CMD, username]

    result = subprocess.run(cmd, capture_output=True, timeout=42, encoding='utf-8', check=False)
    if result.returncode != 0:
        room.backend_data = {
            'timestamp': timezone.now().strftime('%Y-%m-%d %H:%M:%S'),
            '_error': result.stdout,
        }
        room.backend_status = Room.BackendStatus.ERROR
        room.save()

        raise Exception('Backend creation failed.')

    password = result.stdout
    room.backend_link = settings.HANGAR_BACKEND_URL.format(username=username, password=password)
    room.backend_status = Room.BackendStatus.ACTIVE

    all_contacts = []
    all_addrs = set()
    for c in room.assembly.get_all_managing_contacts():
        all_contacts.append(c)
        for a in c.get_all_verified_addresses():
            all_addrs.add(a)

    room.backend_data = {
        'timestamp': timezone.now().strftime('%Y-%m-%d %H:%M:%S'),
        'contacts': list(all_addrs),
    }
    room.save()

    return all_contacts


class Command(BaseCommand):
    def handle(self, *args, **options):
        for x in ['URL', 'BACKEND_URL', 'HOST', 'CMD', 'KEY']:
            if not hasattr(settings, 'HANGAR_' + x) or getattr(settings, 'HANGAR_' + x) in [None, '']:
                raise CommandError(f'No HANGAR_{x} configured.')

        if settings.HANGAR_BACKEND_URL is None or settings.HANGAR_BACKEND_URL == '':
            raise CommandError('No HANGAR_BACKEND_URL configured.')

        rooms = Room.objects.filter(room_type=Room.RoomType.HANGAR, backend_status=Room.BackendStatus.NEW)
        for room in rooms:
            room.backend_status = Room.BackendStatus.SETUP
            room.save()

            try:
                contacts = create_hangar(room)

                msg_subject = f'Hangar "{room.assembly.slug}"'
                msg_text = """Your hangar has been created. Please visit it in the Maschinenraum for access details."""

                for c in contacts:
                    c.notify_user(msg_subject, msg_text, details_link=reverse('backoffice:assembly-room', kwargs={'assembly': room.assembly_id, 'pk': room.id}))

            except Exception as err:
                room.backend_status = Room.BackendStatus.ERROR
                room.backend_data = {
                    **(room.backend_data or {}),
                    'timestamp': timezone.now().strftime('%Y-%m-%d %H:%M:%S'),
                    '_exception': str(err),
                    '_url': room.backend_link,
                }
                room.backend_link = ''
                room.save()
                self.stderr.write(msg=f'Got error on creating hangar: {err}', stylefunc=self.style.ERROR)
