from datetime import timedelta

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db.models import Max

from core.models.activitylog import ActivityLogEntry
from core.models.assemblies import Assembly, AssemblyMember
from core.models.badges import UserBadge
from core.models.conference import (
    ConferenceMember,
    ConferenceTrack,
)
from core.models.dereferrer import UserDereferrerAllowlist
from core.models.events import Event
from core.models.messages import DirectMessage
from core.models.pages import StaticPage
from core.models.schedules import ScheduleSource, ScheduleSourceImport
from core.models.tags import ConferenceTag, TagItem
from core.models.ticket import ConferenceMemberTicket
from core.models.users import PlatformUser, UserCommunicationChannel


class Command(BaseCommand):
    help = 'sanitize database by removing all user-identifying elements'

    def handle(self, *args, **options):
        def print_delete_stat(result, **kwargs):
            if options['verbosity'] >= 2:
                print(result, **kwargs)
            else:
                print(result[0], **kwargs)

        if settings.MOLLY_GUARD:
            print('MOLLY_GUARD is set. Refusing to run. If you really want to run this, unset MOLLY_GUARD.')
            return 'MOLLY_GUARD is set.'

        print('!!!         THIS WILL DESTROY DATA. ARE YOU SURE?           !!!')
        print('!!! Please confirm by typing the Y-word in capital letters. !!!')
        if input() != 'YES':
            print('Aborting.')
            return 'User said "no".'

        users_with_talks = PlatformUser.objects.has_talk_on_conference(conf=None)

        print('AssemblyMember: ', end='', flush=True)
        print_delete_stat(AssemblyMember.objects.all().delete())

        print('Assembly(non-public/non-accepted)')
        non_public_assemblies = Assembly.objects.exclude(state__in=Assembly.PUBLIC_STATES)
        print('  events: ', end='', flush=True)
        print_delete_stat(Event.objects.filter(assembly__in=non_public_assemblies).delete())
        print('  assemblies: ', end='', flush=True)
        print_delete_stat(non_public_assemblies.delete())

        print('Assembly(internal data): ', end='', flush=True)
        non_scrubbed_assemblies = Assembly.objects.exclude(registration_details=None, registration_data=None)
        print(non_scrubbed_assemblies.update(registration_details=None, registration_data=None))

        print('Assembly(technical_user:=None): ', end='', flush=True)
        print(Assembly.objects.exclude(technical_user=None).update(technical_user=None))

        print('ActivityLogEntry: ', end='', flush=True)
        print_delete_stat(ActivityLogEntry.objects.all().delete())

        print('ConferenceMembers: ', end='', flush=True)
        print_delete_stat(ConferenceMember.objects.all().delete())

        print('ConferenceMemberTickets: ', end='', flush=True)
        print_delete_stat(ConferenceMemberTicket.objects.all().delete())

        print('ConferenceTag(non-public): ', end='', flush=True)
        print_delete_stat(ConferenceTag.objects.exclude(is_public=True).delete())

        print('TagItem(missing target): ', end='', flush=True)
        missing_tags = []
        for tag_item in TagItem.objects.all():
            try:
                if tag_item.target is None:
                    missing_tags.append(tag_item.pk)
            except Exception:
                missing_tags.append(tag_item.pk)
        print_delete_stat(TagItem.objects.filter(pk__in=missing_tags).delete())

        print('ConferenceTrack(non-public): ', end='', flush=True)
        Event.objects.filter(track__is_public=False).update(track=None)
        print_delete_stat(ConferenceTrack.objects.exclude(is_public=True).delete())

        print('DirectMessage: ', end='', flush=True)
        print_delete_stat(DirectMessage.objects.all().delete())

        print('Event(non-public): ', end='', flush=True)
        print_delete_stat(Event.objects.filter(is_public=False).delete())

        print('Event(additional_data:={}, owner:=None): ', end='', flush=True)
        print(Event.objects.exclude(additional_data={}, owner=None).update(additional_data={}, owner=None))

        print('UserBadge: ', end='', flush=True)
        print_delete_stat(UserBadge.objects.all().delete())

        print('UserCommunicationChannel: ', end='', flush=True)
        print_delete_stat(UserCommunicationChannel.objects.all().delete())

        print('UserDereferrerAllowlist: ', end='', flush=True)
        print_delete_stat(UserDereferrerAllowlist.objects.all().delete())

        print('PlatformUser (without talks): ', end='', flush=True)
        print_delete_stat(PlatformUser.objects.exclude(pk__in=users_with_talks.values_list('pk', flat=True)).delete())

        print('PlatformUser (with talks, password:=""): ', end='', flush=True)
        print(users_with_talks.update(password=''))

        print('UserCommunicationChannel: ', end='', flush=True)
        print_delete_stat(UserCommunicationChannel.objects.all().delete())

        print('StaticPages: ', end='', flush=True)
        sp_deleted = StaticPage.objects.filter(public_revision=0).delete()[0]
        print(sp_deleted, 'deleted', end=', ', flush=True)
        sp_stripped = 0
        for sp in StaticPage.objects.all():
            sp_stripped += sp.revisions.exclude(revision=sp.public_revision).delete()[0]
        print(sp_stripped, 'revisions stripped')

        print('ScheduleSource(import_frequency:=0): ', end='', flush=True)
        print(ScheduleSource.objects.exclude(import_frequency=timedelta(seconds=0)).update(import_frequency=timedelta(seconds=0)))

        print('ScheduleSourceImports:')
        latest_imports = (
            ScheduleSourceImport.objects.filter(state__in=ScheduleSourceImport.SUCCESS_STATES)
            .values('schedule_source_id', 'schedule_source__assembly__slug')
            .annotate(ts=Max('end'))
        )
        for imp in latest_imports:
            print(' - ', imp['schedule_source__assembly__slug'], ' = ', end='', sep='', flush=True)
            print_delete_stat(ScheduleSourceImport.objects.filter(schedule_source_id=imp['schedule_source_id']).exclude(end=imp['ts']).delete())
