import logging
from typing import TYPE_CHECKING, TypedDict

from pydantic import ConfigDict, TypeAdapter, ValidationError
from rich.console import Console
from rich.markdown import Markdown

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.translation import gettext_lazy as _

if TYPE_CHECKING:
    from core.models.users import PlatformUser

logger = logging.getLogger(__name__)


class ActivityLogEntryManager(models.Manager):
    def for_entity(self, entity):
        ct = ContentType.objects.get_for_model(entity)
        qs = self.get_queryset().filter(entity_content_type=ct, entity_object_id=entity.pk)
        return qs.prefetch_related('user')


class ActivityLogEntry(models.Model):
    class Meta:
        verbose_name = _('ActivityLogEntry')
        verbose_name_plural = _('ActivityLogEntries')
        ordering = ['-timestamp']

    class Kind(models.TextChoices):
        SYSTEM = 'sys', _('ActivityLogEntry__kind-sys')
        """System-triggered log entry, i.e. schedule import"""
        MESSAGE = 'message', _('ActivityLogEntry__kind-message')
        """Message sent to the entity from an orga member"""
        STAFF = 'staff', _('ActivityLogEntry__kind-staff')
        """Logged activity by a staff member."""
        ENTITY = 'entity', _('ActivityLogEntry__kind-entity')
        """Logged activity by an entity member."""
        OTHER = 'other', _('ActivityLogEntry__kind-other')
        """Other kind of activity log entry. (e.g., users external to the entity)"""

    objects = ActivityLogEntryManager()

    entity_content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    entity_object_id = models.UUIDField()
    entity = GenericForeignKey(
        'entity_content_type',
        'entity_object_id',
    )

    user = models.ForeignKey(
        'PlatformUser',
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        help_text=_('ActivityLogEntry__user__help'),
        verbose_name=_('ActivityLogEntry__user'),
    )
    """The PlatformUser responsible for this activity, can be null for system-induced changes (or if the user got deleted)."""

    timestamp = models.DateTimeField(
        auto_now_add=True,
        editable=False,
        help_text=_('ActivityLogEntry__timestamp__help'),
        verbose_name=_('ActivityLogEntry__timestamp'),
    )
    """When has this change happened."""

    kind = models.CharField(
        max_length=20,
        choices=Kind.choices,
        help_text=_('ActivityLogEntry__kind__help'),
        verbose_name=_('ActivityLogEntry__kind'),
    )
    """What type of change is this, i.e. user-induced, automatic by system, etc."""

    context = models.CharField(
        max_length=100,
        blank=True,
        editable=False,
        null=True,
        help_text=_('ActivityLogEntry__context__help'),
        verbose_name=_('ActivityLogEntry__context'),
    )
    """From where was this triggered, i.e. UI, API, Backoffice, etc."""

    changes = models.JSONField(
        blank=True,
        null=True,
        help_text=_('ActivityLogEntry__changes__help'),
        verbose_name=_('ActivityLogEntry__changes'),
    )
    """The changed fields as a dict with values being a tuple of (old_value, new_value)."""

    comment = models.TextField(
        blank=True,
        null=True,
        help_text=_('ActivityLogEntry__comment__help'),
        verbose_name=_('ActivityLogEntry__comment'),
    )
    """A free-text comment, likely to be added by staff in the backoffice."""

    patrolled = models.BooleanField(
        null=True,
        default=False,
        help_text=_('ActivityLogEntry__patrolled__help'),
        verbose_name=_('ActivityLogEntry__patrolled'),
    )
    """Flag if this change has been reviewed. Used for moderation visibility purposes only, no functional influence."""

    @property
    def conference(self):
        if (entity := self.entity) is None:
            return None
        return entity.conference

    def __str__(self):
        return f'{self.timestamp} [{self.kind}]' + (f' @{self.user.username}' if self.user else '') + (f' "{self.comment}"' if self.comment else '')


class ActivityLogChange(TypedDict, total=False):
    __pydantic_config__ = ConfigDict(extra='forbid')

    new: str
    old: str


ActivityLogChangeValidator = TypeAdapter(ActivityLogChange)


class ActivityLogMixin:
    @property
    def logentries(self):
        return ActivityLogEntry.objects.for_entity(entity=self)

    def get_activity_log_kind_for_user(self, user: 'PlatformUser') -> ActivityLogEntry.Kind:
        raise ImproperlyConfigured('get_log_kind_for_user() must be implemented in the subclass.')

    def log_activity(
        self,
        user: 'PlatformUser | None',
        kind: str | None = None,
        context: str | None = None,
        comment: str | None = None,
        **changes: str | ActivityLogChange,
    ):
        # sanity-check changes values
        if changes:
            for k, v in changes.items():
                if isinstance(v, dict):
                    try:
                        ActivityLogChangeValidator.validate_python(v)
                        if 'new' in v and v['new'] == '':
                            del v['new']
                        if 'old' in v and v['old'] == '':
                            del v['old']
                    except ValidationError as exc:
                        raise ValueError(
                            "ActivityLogEntry's changes must be of form field={'old': old_value, 'new': new_value} "
                            f"or field=initial_value, '{k}' does not conform."
                        ) from exc
                elif isinstance(v, str):
                    pass
                else:
                    raise ValueError(
                        "ActivityLogEntry's changes must be of form field={'old': old_value, 'new': new_value} "
                        f"or field=initial_value, '{k}' does not conform."
                    )

        # skip empty activity
        if not changes and not comment:
            logging.debug('Skipping empty log activity on <%s> by <%s> (%s).', self, user, kind)
            return

        if not kind:
            if not user:
                raise ValueError('kind must be provided if user is not.')
            kind = self.get_activity_log_kind_for_user(user)

        # create the entry
        ActivityLogEntry.objects.create(
            entity=self,
            user=user,
            kind=kind,
            context=context,
            comment=comment,
            changes=changes,
        )

        if logger.isEnabledFor(logging.TRACE):
            # log the changes in a human-readable format
            # We need to output this via console as the logger does not support rich Markdown text
            console = Console()
            md = f'**DEBUG INFO**: *changes* on: {self}\n\n'
            md += '| field | value |\n'
            md += '| --- | --- |\n'

            for k, v in changes.items():
                if isinstance(v, dict):
                    md += f'| {k} | '
                    if 'old' in v:
                        md += f'~~{v["old"]}~~ '
                    if 'new' in v:
                        md += f'{v["new"]}'
                    md += ' |\n'
                else:
                    md += f'| {k} | {v} |\n'
            console.print(Markdown(md))


def model_form_valid_activitylog(
    context: str | None = None,
    kind: 'ActivityLogEntry.Kind' = ActivityLogEntry.Kind.ENTITY,
):
    """
    A decorator to be used on FormMixin.form_valid() function to log the activity.
    This stores all the form fields' values prior to calling the original form_valid() function and
    later gets the current values, comparing those.
    With those changes, a new ActivityLogEntry is being created.
    :param kind: type of log entry, see ActivityLogEntry.Kind enum - defaults to 'ENTITY'
    :param context: an optional string to be logged as the context of the ActivityLogEntry
    :return: the decorated function
    """

    def decorator(form_valid_func):
        def wrapper(self, form, *args, **kwargs):
            # backup current values, if possible
            old_values = {fld_name: form.get_initial_for_field(field, fld_name) for fld_name, field in form.fields.items()}

            # CALL the original form_valid(), logging possible exception
            r = form_valid_func(self, form, *args, **kwargs)

            # note down changed model fields
            changes = {}
            for field in form.changed_data:
                if hasattr(form.instance, field):
                    # model field
                    new_value = getattr(form.instance, field)
                else:
                    # value in form only (e.g. additional field)
                    new_value = form[field]
                changes[field] = (old_values[field], new_value) if field in old_values else new_value

            # save the update data
            if form.instance is None or form.instance.pk is None:
                log_changes = {k: (str(v[1]) if isinstance(v, tuple) else str(v)) for k, v in changes.items()}
            else:
                log_changes = {
                    k: (
                        ActivityLogChange(
                            old=str(v[0]) if v[0] is not None else '',
                            new=str(v[1]),
                        )
                        if isinstance(v, tuple)
                        else str(v)
                    )
                    for k, v in changes.items()
                }
            form.instance.log_activity(
                user=self.request.user if hasattr(self, 'request') else None,
                kind=kind,
                context=context,
                **log_changes,
            )

            return r

        return wrapper

    return decorator
