import string
from typing import Any

from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import Page, Paginator
from django.db.models.base import Model
from django.db.models.query import QuerySet


class AlphabetPaginator(Paginator):
    """Pagination for string-based objects"""

    def __init__(self, *args, on: str | None = None, **kwargs) -> None:
        if on is None:
            raise ImproperlyConfigured('on parameter required for NamePaginator')
        self.on = on
        super().__init__(*args, **kwargs)
        self.item_dict = {}

        for item in self.object_list:
            label = getattr(item, self.on).lower()
            letter = label[0] if label[0] in string.ascii_lowercase else '0'
            self.item_dict.setdefault(letter, []).append(item)

    def __iter__(self):
        for page_index in ['0', *string.ascii_lowercase]:
            yield self.page(page_index)

    def get_page(self, index: int | float | str | None) -> Page:
        if index.lower() not in ['0', *string.ascii_lowercase]:
            index = '0'

        return self.page(index)

    def _get_page(self, *args, **kwargs):
        return LetterPage(*args, **kwargs)

    def page(self, char: str | int):
        """Returns a Page object for the given 1-based page number."""
        lower_char = str(char).lower()

        # Allow 1 to account for defaults in ListView

        return self._get_page(self.item_dict.get(char, []), lower_char, self)

    @property
    def num_pages(self):
        """Returns the total number of pages"""
        return len(string.ascii_uppercase) + 1


class LetterPage(Page):
    def __init__(self, object_list: list[dict[str, str]] | list[Model] | list[int] | QuerySet[Any] | str, index: int, paginator: Paginator) -> None:
        number = 1 if index == '0' else ord(index) + 2 - ord('a')
        self.index = index
        super().__init__(object_list, number, paginator)

    def next_page_number(self):
        raise NotImplementedError('Use next_page_index')

    def previous_page_number(self):
        raise NotImplementedError('Use previous_page_index')

    def next_page_index(self):
        return chr(ord('a') + self.number - 1)

    def previous_page_index(self):
        return chr(ord('a') + self.number - 3) if self.number > 2 else '0'
