from datetime import timedelta

from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils.timezone import now

from core.models.assemblies import Assembly
from core.models.conference import Conference, ConferenceMember
from core.models.events import Event
from core.models.rooms import Room
from core.models.schedules import ScheduleSource, ScheduleSourceMapping
from core.models.users import PlatformUser


class EventTests(TestCase):
    def setUp(self):
        self.conference1 = Conference(slug='foo', name='Foo Conference', start=now() - timedelta(days=1), end=now() + timedelta(days=1))
        self.conference1.save()
        self.conference2 = Conference(slug='bar', name='Bar Conference')
        self.conference2.save()

        self.user = PlatformUser(username='bernd')
        self.user.save()
        ConferenceMember(conference=self.conference1, user=self.user).save()
        ConferenceMember(conference=self.conference2, user=self.user).save()

        self.assembly1a = Assembly(conference=self.conference1, slug='fnord', name='Fnord Assembly', is_official=True)
        self.assembly1a.save()
        self.assembly1b = Assembly(conference=self.conference1, slug='banana', name='Obstgarten')
        self.assembly1b.save()
        self.assembly2a = Assembly(conference=self.conference2, slug='fnord', name='Fnord Assembly', is_official=True)
        self.assembly2a.save()
        self.assembly2b = Assembly(conference=self.conference2, slug='banana', name='Obstgarten')
        self.assembly2b.save()

        self.room1a = Room(conference=self.conference1, assembly=self.assembly1a, name='Saal 1', room_type=Room.RoomType.LECTURE_HALL)
        self.room1a.save()
        self.room1b = Room(conference=self.conference1, assembly=self.assembly1b, name='Apfelbaum', room_type=Room.RoomType.WORKSHOP)
        self.room1b.save()
        self.room2a = Room(conference=self.conference2, assembly=self.assembly2a, name='Ada', room_type=Room.RoomType.LECTURE_HALL)
        self.room2a.save()
        self.room2a = Room(conference=self.conference2, assembly=self.assembly2a, name='Brombeerbusch', room_type=Room.RoomType.WORKSHOP)
        self.room2a.save()

        self.valid_event = {
            'conference': self.conference1,
            'assembly': self.assembly1a,
            'name': 'testEvent',
            'slug': 'testEvent',
            'kind': Event.Kind.ASSEMBLY,
            'room': self.room1a,
            'schedule_start': now(),
            'schedule_duration': timedelta(hours=1),
            'is_public': True,
        }

    def test_creation(self):
        e1 = Event(conference=self.conference1, assembly=self.assembly1a, name='Opening', kind=Event.Kind.OFFICIAL, room=self.room1a)
        e1.save()
        e2 = Event(conference=self.conference1, assembly=self.assembly1b, name='Example', kind=Event.Kind.ASSEMBLY)
        e2.save()

        self.assertFalse(e1.is_imported)
        self.assertFalse(e2.is_imported)

    def test_creation_from_dict(self):
        event = Event.objects.create(**self.valid_event)
        self.assertEqual(event.name, self.valid_event['name'])
        self.assertEqual(event.schedule_end, self.valid_event['schedule_start'] + self.valid_event['schedule_duration'])

    def test_error_assembly_wrong_conference(self):
        event_data = self.valid_event.copy()
        event_data['assembly'] = self.assembly2a
        with self.assertRaises(ValidationError):
            event = Event(**event_data)
            event.full_clean()

    def test_error_negative_duration(self):
        event_data = self.valid_event.copy()
        event_data['schedule_duration'] = timedelta(-1)
        with self.assertRaises(ValidationError) as cm:
            event = Event(**event_data)
            event.full_clean()
        self.assertTrue('schedule_duration' in cm.exception.error_dict)

    def test_error_zero_duration(self):
        event_data = self.valid_event.copy()
        event_data['schedule_duration'] = timedelta(0)
        with self.assertRaises(ValidationError) as cm:
            event = Event(**event_data)
            event.full_clean()
        self.assertTrue('schedule_duration' in cm.exception.error_dict)

    def test_error_end_after_conference(self):
        event_data = self.valid_event.copy()
        event_data['schedule_duration'] = timedelta(days=2)
        with self.assertRaises(ValidationError) as cm:
            event = Event(**event_data)
            event.full_clean()
        self.assertTrue('schedule_duration' in cm.exception.error_dict)

    def test_error_start_before_conference(self):
        event_data = self.valid_event.copy()
        event_data['schedule_start'] = event_data['schedule_start'] - timedelta(days=2)
        with self.assertRaises(ValidationError) as cm:
            event = Event(**event_data)
            event.full_clean()
        self.assertTrue('schedule_start' in cm.exception.error_dict)

    def test_no_sos_assembly(self):
        event_data = self.valid_event.copy()
        event_data['kind'] = Event.Kind.SELF_ORGANIZED
        with self.assertRaises(ValidationError):
            event = Event(**event_data)
            event.full_clean()

    def test_imported(self):
        e = Event(conference=self.conference1, assembly=self.assembly1a, name='Imported', kind=Event.Kind.ASSEMBLY)
        e.save()

        # fake an existing import mapping
        src = ScheduleSource.objects.create(conference=self.conference1)
        src.mappings.create(mapping_type=ScheduleSourceMapping.MappingType.EVENT, source_id='unittest', local_id=e.pk, skip=False)

        # the event's property should pick this up
        self.assertTrue(e.is_imported)

    def test_clean(self):
        with self.assertRaises(ValidationError):
            e = Event(conference=self.conference1, assembly=self.assembly1b, name='official in nonofficial assembly', kind=Event.Kind.OFFICIAL)
            e.full_clean()
        with self.assertRaises(ValidationError):
            e = Event(conference=self.conference1, assembly=self.assembly2a, name='assembly in another conference', kind=Event.Kind.OFFICIAL)
            e.full_clean()
        with self.assertRaises(ValidationError):
            e = Event(conference=self.conference1, assembly=self.assembly1b, name='room of another assembly', room=self.room1a, kind=Event.Kind.OFFICIAL)
            e.full_clean()

    def test_dict(self):
        # create a event from data
        current_date = now()
        data = {'name': 'foo', 'schedule_start': current_date.isoformat()}
        e = Event.from_dict(data, conference=self.conference1, assembly=self.assembly1a)
        self.assertEqual(e.name, data['name'])
        self.assertIsNotNone(e.schedule_start)
        self.assertIsNone(e.schedule_end)
        e.save()

        # update that event with an end date
        data = {'schedule_end': (current_date + timedelta(hours=10)).isoformat()}
        e = Event.from_dict(data, conference=self.conference1, assembly=self.assembly1a, existing=e)
        e.save()
        self.assertIsNotNone(e.schedule_end)
        self.assertEqual(e.schedule_duration, timedelta(hours=10))

        # ensure kind cannot be set (unless unallowed)
        e = Event.from_dict({'kind': 'official'}, conference=self.conference1, assembly=self.assembly1a, allow_kind=False)
        self.assertEqual(e.kind, Event.Kind.ASSEMBLY)
        e = Event.from_dict({'kind': 'official'}, conference=self.conference1, assembly=self.assembly1a, allow_kind=True)
        self.assertEqual(e.kind, Event.Kind.OFFICIAL)

    def _expect_clean_validation_error(self, obj, field, **params):
        with self.assertRaises(ValidationError) as ve:
            obj.full_clean()
        self.assertEqual(1, len(ve.exception.error_dict), 'Expected only one validation error.')
        self.assertIn(field, ve.exception.error_dict, f'Expected a validation error on "{field}".')
        self.assertEqual(1, len(ve.exception.error_dict[field]), f'Expected only one validation error for "{field}".')
        ve_field = ve.exception.error_dict[field][0]
        for k, v in params.items():
            self.assertIn(k, ve_field.params, f'Expected the ValidationError on "{field}" to have a param named "{k}".')
            if isinstance(v, list):
                self.assertListEqual(v, ve_field.params[k])
            else:
                self.assertEqual(v, ve_field.params[k])

    def test_collision(self):
        # create blocker event (kind=official, is_public=False) from now..now+2h
        blocker = Event(
            conference=self.conference1,
            assembly=self.assembly1a,
            name='BLOCKER',
            slug='blocker',
            kind=Event.Kind.OFFICIAL,
            is_public=False,
            schedule_start=now(),
            schedule_duration=timedelta(hours=2),
        )
        blocker.room = self.room1a
        blocker.save()

        # create an event before the blocker with a 15 Min. gap
        before = Event(
            conference=self.conference1,
            assembly=self.assembly1a,
            name='before',
            slug='before',
            kind=Event.Kind.ASSEMBLY,
            is_public=True,
            schedule_start=now() - timedelta(hours=1),
            schedule_duration=timedelta(minutes=45),
        )
        before.room = self.room1a
        before.save()

        # attempt creation of a colliding event (starts during "before" and ends in "blocker")
        fail = Event(
            conference=self.conference1,
            assembly=self.assembly1a,
            name='fail',
            slug='fail',
            kind=Event.Kind.ASSEMBLY,
            is_public=True,
            schedule_start=now() - timedelta(minutes=30),
            schedule_duration=timedelta(minutes=60),
        )
        fail.room = self.room1a
        self._expect_clean_validation_error(fail, 'room', hidden=1, slugs=['before'])

        # sneak a short event in the gap
        sneak = Event(
            conference=self.conference1,
            assembly=self.assembly1a,
            name='sneak',
            slug='sneak',
            kind=Event.Kind.ASSEMBLY,
            is_public=True,
            schedule_start=now() - timedelta(minutes=10),
            schedule_duration=timedelta(minutes=5),
        )
        sneak.room = self.room1a
        sneak.save()

        # append an extra event after the blocker
        closure = Event(
            conference=self.conference1,
            assembly=self.assembly1a,
            name='closure',
            slug='closure',
            kind=Event.Kind.ASSEMBLY,
            is_public=True,
            schedule_start=now() + timedelta(hours=2, minutes=15),
            schedule_duration=timedelta(minutes=30),
        )
        closure.room = self.room1a
        closure.save()

        # attempt the colliding event again, should now see the "sneak" event (but not the one after the blocker)
        self._expect_clean_validation_error(fail, 'room', hidden=1, slugs=['before', 'sneak'])

        # extending the duration should include the "closure" event, too
        fail.schedule_duration = timedelta(hours=3)
        self._expect_clean_validation_error(fail, 'room', hidden=1, slugs=['before', 'sneak', 'closure'])

        # moving into the closure event (extending over the end) should only show that one
        fail.schedule_start = now() + timedelta(hours=2, minutes=30)
        self._expect_clean_validation_error(fail, 'room', hidden=0, slugs=['closure'])
