import logging
import string
from typing import Any

from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import InvalidPage, Paginator
from django.db.models import Count, Prefetch, QuerySet
from django.http import Http404
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView

from core.models.tags import TagItem
from core.paginators import AlphabetPaginator

logger = logging.getLogger(__name__)


class FilteredListView(ListView):
    filters: dict[str, tuple[str, str]] | None = None
    params: dict[str, str]

    def __init__(self, **kwargs: Any) -> None:
        self.params = {}
        if self.filters is None:
            msg = 'View Filters are required'
            raise ImproperlyConfigured(msg)
        super().__init__(**kwargs)

    def get_paginate_by(self, queryset=None):
        """
        Paginate by specified value in querystring, or use default class property value.
        """
        try:
            paginate_by = self.request.GET.get('paginate_by', self.paginate_by)
            if paginate_by is None:
                return self.paginate_by
            paginate_by = int(paginate_by)
        except ValueError:
            if str(paginate_by).lower() == 'all':
                paginate_by = False
            else:
                current_messages = [message.message for message in messages.get_messages(self.request)]
                message = _('Invalid paginate_by value, using default')
                if message not in list(current_messages):
                    messages.error(self.request, message)
                paginate_by = self.paginate_by
        return paginate_by

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        paginate_by = self.get_paginate_by()
        return {
            **super().get_context_data(**kwargs),
            'params': {
                **self.params,
                'paginate_by': paginate_by if paginate_by is not False else _('all'),
            },
        }

    def get_queryset(self, queryset: QuerySet | None = None, query_filters: dict[str, str | list | bool] | None = None) -> QuerySet[Any]:
        queryset = super().get_queryset() if queryset is None else queryset
        ordering = self.get_ordering()
        if ordering:
            if isinstance(ordering, str):
                ordering = (ordering,)
            queryset = queryset.order_by(*ordering)
        query_filters = query_filters or {}
        get_parameters = getattr(self.request, 'GET', {})
        for parameter, (query_parameter, parameter_type) in self.filters.items():
            value: str | bool | list[str]
            if value := get_parameters.get(parameter, self.kwargs.get(parameter)):
                self.params[parameter] = str(value)
                match parameter_type:
                    case 'list':
                        value = value.split(',')
                        value = [x.strip() for x in value]
                        query_filters[f'{query_parameter}__in'] = value
                    case 'bool':
                        value = value == 'true'
                        query_filters[query_parameter] = value
                    case _:
                        query_filters[query_parameter] = value
        logger.trace(f'Query Filters: {query_filters}')
        for key, value in query_filters.items():
            match key[0]:
                case '!':
                    queryset = queryset.exclude(**{key[1:]: value})
                case 'X':
                    queryset = queryset.exclude(value)
                case 'Q':
                    queryset = queryset.filter(value)
                case _:
                    queryset = queryset.filter(**{key: value})
        return queryset


class AlphabeticalPaginatorView(FilteredListView):
    full_queryset: bool = False

    def get_paginate_by(self, queryset=None):
        return True

    def get_paginator(self, queryset: QuerySet[Any], per_page: int, orphans: int = ..., allow_empty_first_page: bool = ..., **kwargs: Any) -> Paginator:
        return AlphabetPaginator(queryset, per_page, orphans, allow_empty_first_page, on='name', **kwargs)

    def paginate_queryset(self, queryset, page_size):
        """Paginate the queryset, if needed."""
        paginator = self.get_paginator(
            queryset,
            page_size,
            orphans=self.get_paginate_orphans(),
            allow_empty_first_page=self.get_allow_empty(),
        )
        page_kwarg = self.page_kwarg
        page_index = self.kwargs.get(page_kwarg) or self.request.GET.get(page_kwarg)
        if page_index is None:
            self.full_queryset = True
            page_index = '0'
        page_index = str(page_index).lower()

        if str(page_index) not in ['0', *string.ascii_lowercase]:
            Http404(_('Invalid alphabetical page'))

        try:
            page = paginator.page(page_index)
            return (paginator, page, page.object_list, page.has_other_pages())
        except InvalidPage as e:
            raise Http404(_('Invalid page (%(page_number)s): %(message)s') % {'page_number': page_index, 'message': str(e)})

    def get_context_data(self, object_list=None, **kwargs: Any) -> dict[str, Any]:
        queryset = object_list if object_list is not None else self.object_list
        context = super().get_context_data(**kwargs)
        context_object_name = self.get_context_object_name(queryset)
        del context['params']['paginate_by']
        if self.full_queryset:
            context['view_all'] = True
            context['object_list'] = queryset
            if context_object_name is not None:
                context[context_object_name] = queryset
        return context


class PrefetchTagsMixin(FilteredListView):
    def add_tag_annotations(self, queryset: QuerySet) -> QuerySet[Any]:
        return queryset.annotate(Count('tags', distinct=True)).prefetch_related(
            Prefetch('tags', to_attr='prefetched_tags', queryset=TagItem.objects.select_related('tag').filter(tag__is_public=True)),
        )
