import logging
import re
from difflib import HtmlDiff

from django.conf import settings
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import get_language

from core.markdown import render_markdown_ex
from core.models import Conference, StaticPage

from .forms import ReportForm

logger = logging.getLogger(__name__)


def fetch_wiki_page(
    conference: Conference, slug: str, language: str | None = None, default_title: str | None = 'Page 404', default_body: str | None = None
) -> tuple[str, str]:
    """
    Retrieves the wiki page with the given slug in the given language (defaults to current request's one).
    :param conference: the Conference in which to search
    :param slug: the page's slug, see StaticPage.slug
    :param language: the requested language (if the targeted wiki page is multilingual), if not given defaults to the current request's one
    :param default_title: the title to be returned if the page does not exist
    :param default_body: the body markup to use if the page does not exist (a default warning is provided)
    :return: a tuple (title, body_html) for the requested page
    """
    try:
        static_page = StaticPage.objects.conference_accessible(conference=conference, language=language or get_language()).get(slug=slug)

    except StaticPage.DoesNotExist:
        url = reverse('plainui:static_page_edit', kwargs={'page_slug': slug})
        if not default_title:
            default_title = 'Page Missing'

        if not default_body:
            default_body = '!!! warning\n  WIKI PAGE 404 `{slug}` [(edit)]({url})'

        static_page = StaticPage(conference=conference, title=default_title, body=default_body.format(slug=slug, url=url))
        static_page.body_html = render_markdown_ex(conference, static_page.body).document

    return static_page.title, mark_safe(static_page.body_html)


# TODO: Implement or remove after discussion
def check_message_content(conf, request, text, kind, kind_data):
    try:
        trigger_autoban = False
        for pattern in settings.AUTOBAN_KEYWORDS:
            if isinstance(pattern, re.Pattern):
                trigger_autoban |= pattern.search(text) is not None
            elif pattern in text:
                trigger_autoban = True

        if not trigger_autoban:
            return False

        request.user.shadow_banned = True
        request.user.save(update_fields=['shadow_banned'])

        report_form = ReportForm(
            conf=conf,
            request=request,
            data={
                'kind': kind,
                'kind_data': str(kind_data),
                'category': 'abuse',
                'message': 'Autoban triggered',
                'message2': '.',
            },
        )
        report_form.is_valid()

        report_form.send_report_mail(request)
        return True
    except Exception:
        raise Exception('Boom')  # don't allow leaking eg. ValidationErrors that might be handled in a View


class StaticPageDiff(HtmlDiff):
    # function from source code - removed `nowrap="nowrap"`
    def _format_line(self, side, flag, linenum, text):
        try:
            linenum = f'{linenum:d}'
            change_id = f' id="{self._prefix[side]}{linenum}"'
        except ValueError:
            # handle blank lines where linenum is '>' or ''
            change_id = ''

        # replace those things that would get confused with HTML symbols
        text = text.replace('&', '&amp;').replace('>', '&gt;').replace('<', '&lt;')

        # make space non-breakable so they don't get compressed or line wrapped
        text = text.replace(' ', '&nbsp;').rstrip()

        # vvv ---- removed `nowrap="nowrap"` --- vvv
        return f'<td class="diff_header"{change_id}>{linenum}</td><td class="diff_content">{text}</td>'
